@@ -8,6 +8,7 @@
 #include <algorithm>
 #include "GLSLANG/ShaderLang.h"
 #include "CanvasUtils.h"
+#include "gfxEnv.h"
 #include "GLContext.h"
 #include "jsfriendapi.h"
 #include "mozilla/CheckedInt.h"
@@ -285,21 +286,6 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
   mCanLoseContextInForeground =
       StaticPrefs::webgl_can_lose_context_in_foreground();
 
-  // These are the default values, see 6.2 State tables in the
-  // OpenGL ES 2.0.25 spec.
-  mDriverColorMask = mColorWriteMask;
-  mColorClearValue[0] = 0.f;
-  mColorClearValue[1] = 0.f;
-  mColorClearValue[2] = 0.f;
-  mColorClearValue[3] = 0.f;
-  mDepthWriteMask = true;
-  mDepthClearValue = 1.f;
-  mStencilClearValue = 0;
-  mStencilRefFront = 0;
-  mStencilRefBack = 0;
-
-  mLineWidth = 1.0;
-
   /*
   // Technically, we should be setting mStencil[...] values to
   // `allOnes`, but either ANGLE breaks or the SGX540s on Try break.
@@ -326,23 +312,7 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
   AssertUintParamCorrect(gl, LOCAL_GL_STENCIL_BACK_WRITEMASK,
                          mStencilWriteMaskBack);
 
-  mDitherEnabled = true;
-  mRasterizerDiscardEnabled = false;
-  mScissorTestEnabled = false;
-
-  mDepthTestEnabled = 0;
-  mDriverDepthTest = false;
-  mStencilTestEnabled = 0;
-  mDriverStencilTest = false;
-
-  mGenerateMipmapHint = LOCAL_GL_DONT_CARE;
-
   // Bindings, etc.
-  mActiveTexture = 0;
-  mDefaultFB_DrawBuffer0 = LOCAL_GL_BACK;
-  mDefaultFB_ReadBuffer = LOCAL_GL_BACK;
-
-  mWebGLError = LOCAL_GL_NO_ERROR;
 
   mBound2DTextures.Clear();
   mBoundCubeMapTextures.Clear();
@@ -560,12 +530,6 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     return false;
   }
 
-  if (!gl->IsSupported(gl::GLFeature::vertex_array_object)) {
-    *out_failReason = {"FEATURE_FAILURE_WEBGL_VAOS",
-                       "Requires vertex_array_object."};
-    return false;
-  }
-
   // OpenGL core profiles remove the default VAO object from version
   // 4.0.0. We create a default VAO for all core profiles,
   // regardless of version.
@@ -579,6 +543,8 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
 
   mPrimRestartTypeBytes = 0;
 
+  // -
+
   mGenericVertexAttribTypes.assign(limits.maxVertexAttribs,
                                    webgl::AttribBaseType::Float);
   mGenericVertexAttribTypeInvalidator.InvalidateCaches();
@@ -589,6 +555,33 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
 
   mFakeVertexAttrib0BufferObject = 0;
 
+  mNeedsLegacyVertexAttrib0Handling = gl->IsCompatibilityProfile();
+  if (gl->WorkAroundDriverBugs() && kIsMacOS) {
+    // Failures in conformance/attribs/gl-disabled-vertex-attrib.
+    // Even in Core profiles on NV. Sigh.
+    mNeedsLegacyVertexAttrib0Handling |= (gl->Vendor() == gl::GLVendor::NVIDIA);
+
+    mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst |=
+        (gl->Vendor() == gl::GLVendor::Intel);
+
+    // Failures for programs with no attribs:
+    // conformance/attribs/gl-vertex-attrib-unconsumed-out-of-bounds.html
+    mMaybeNeedsLegacyVertexAttrib0Handling = true;
+  }
+  mMaybeNeedsLegacyVertexAttrib0Handling |= mNeedsLegacyVertexAttrib0Handling;
+
+  if (const auto& env =
+          PR_GetEnv("MOZ_WEBGL_WORKAROUND_FIRST_AFFECTS_INSTANCE_ID")) {
+    const auto was = mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst;
+    mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst =
+        (std::string{env} != "0");
+    printf_stderr(
+        "mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst: %i -> %i\n",
+        int(was), int(mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst));
+  }
+
+  // -
+
   mNeedsIndexValidation =
       !gl->IsSupported(gl::GLFeature::robust_buffer_access_behavior);
   switch (StaticPrefs::webgl_force_index_validation()) {
