@@ -7,30 +7,38 @@
 #define WEBGLTYPES_H_
 
 #include <limits>
+#include <string>
 #include <type_traits>
 #include <unordered_map>
 #include <vector>
 
-// Most WebIDL typedefs are identical to their OpenGL counterparts.
 #include "GLDefs.h"
+#include "ImageContainer.h"
 #include "mozilla/Casting.h"
 #include "mozilla/CheckedInt.h"
+#include "mozilla/MathAlgorithms.h"
 #include "mozilla/Range.h"
 #include "mozilla/RefCounted.h"
+#include "mozilla/Result.h"
+#include "mozilla/ResultVariant.h"
+#include "mozilla/gfx/2D.h"
+#include "mozilla/gfx/BuildConstants.h"
 #include "mozilla/gfx/Point.h"
+#include "mozilla/gfx/Rect.h"
+#include "mozilla/ipc/Shmem.h"
+#include "mozilla/layers/LayersSurfaces.h"
 #include "gfxTypes.h"
 
 #include "nsTArray.h"
 #include "nsString.h"
 #include "mozilla/dom/WebGLRenderingContextBinding.h"
 #include "mozilla/ipc/SharedMemoryBasic.h"
-//#include "WebGLStrongTypes.h"
 
 // Manual reflection of WebIDL typedefs that are different from their
 // OpenGL counterparts.
-typedef int64_t WebGLsizeiptr;
-typedef int64_t WebGLintptr;
-typedef bool WebGLboolean;
+using WebGLsizeiptr = int64_t;
+using WebGLintptr = int64_t;
+using WebGLboolean = bool;
 
 // -
 
@@ -94,6 +102,9 @@ inline auto AutoAssertCast(const From val) {
   return detail::AutoAssertCastT<From>(val);
 }
 
+const char* GetEnumName(GLenum val, const char* defaultRet = "<unknown>");
+std::string EnumString(GLenum val);
+
 namespace webgl {
 template <typename T>
 struct QueueParamTraits;
@@ -220,7 +231,9 @@ enum class WebGLExtensionID : uint8_t {
   EXT_texture_compression_bptc,
   EXT_texture_compression_rgtc,
   EXT_texture_filter_anisotropic,
+  EXT_texture_norm16,
   MOZ_debug,
+  OES_draw_buffers_indexed,
   OES_element_index_uint,
   OES_fbo_render_mipmap,
   OES_standard_derivatives,
@@ -246,37 +259,27 @@ enum class WebGLExtensionID : uint8_t {
   Max
 };
 
-template <typename T>
-inline constexpr auto EnumValue(const T v) {
-  return static_cast<typename std::underlying_type<T>::type>(v);
-}
-
-class UniqueBuffer {
+class UniqueBuffer final {
   // Like UniquePtr<>, but for void* and malloc/calloc/free.
-  void* mBuffer;
+  void* mBuffer = nullptr;
 
  public:
-  UniqueBuffer() : mBuffer(nullptr) {}
-
-  MOZ_IMPLICIT UniqueBuffer(void* buffer) : mBuffer(buffer) {}
-
-  ~UniqueBuffer() { free(mBuffer); }
-
-  UniqueBuffer(UniqueBuffer&& other) {
-    this->mBuffer = other.mBuffer;
-    other.mBuffer = nullptr;
+  static inline UniqueBuffer Take(void* buffer) {
+    UniqueBuffer ret;
+    ret.mBuffer = buffer;
+    return ret;
   }
 
-  UniqueBuffer& operator=(UniqueBuffer&& other) {
-    free(this->mBuffer);
-    this->mBuffer = other.mBuffer;
-    other.mBuffer = nullptr;
-    return *this;
-  }
+  UniqueBuffer() = default;
 
-  UniqueBuffer& operator=(void* newBuffer) {
-    free(this->mBuffer);
-    this->mBuffer = newBuffer;
+  ~UniqueBuffer() { reset(); }
+
+  UniqueBuffer(UniqueBuffer&& rhs) { *this = std::move(rhs); }
+
+  UniqueBuffer& operator=(UniqueBuffer&& rhs) {
+    reset();
+    this->mBuffer = rhs.mBuffer;
+    rhs.mBuffer = nullptr;
     return *this;
   }
 
@@ -284,15 +287,22 @@ class UniqueBuffer {
 
   void* get() const { return mBuffer; }
 
-  UniqueBuffer(const UniqueBuffer& other) =
-      delete;  // construct using std::move()!
-  void operator=(const UniqueBuffer& other) =
-      delete;  // assign using std::move()!
+  void reset() {
+    // Believe it or not, when `free` unconditional, it was showing up
+    // in profiles, nearly 20% of time spent in MethodDispatcther<UniformData>
+    // on Aquarium.
+    if (mBuffer) {
+      free(mBuffer);
+      mBuffer = nullptr;
+    }
+  }
 };
 
 namespace webgl {
 struct FormatUsageInfo;
 
+static constexpr GLenum kErrorPerfWarning = 0x10001;
+
 struct SampleableInfo final {
   const char* incompleteReason = nullptr;
   uint32_t levels = 0;
@@ -318,7 +328,7 @@ enum class UniformBaseType : uint8_t {
 };
 const char* ToString(UniformBaseType);
 
-typedef uint64_t ObjectId;
+using ObjectId = uint64_t;
 
 enum class BufferKind : uint8_t {
   Undefined,
@@ -332,40 +342,15 @@ enum class BufferKind : uint8_t {
 
 struct FloatOrInt final  // For TexParameter[fi] and friends.
 {
-  const bool isFloat;
-  const GLfloat f;
-  const GLint i;
+  bool isFloat = false;
+  GLfloat f = 0;
+  GLint i = 0;
 
   explicit FloatOrInt(GLint x = 0) : isFloat(false), f(x), i(x) {}
 
   explicit FloatOrInt(GLfloat x) : isFloat(true), f(x), i(roundf(x)) {}
-
-  FloatOrInt& operator=(const FloatOrInt& x) {
-    memcpy(this, &x, sizeof(x));
-    return *this;
-  }
 };
 
-struct WebGLPixelStore final {
-  uint32_t mUnpackImageHeight = 0;
-  uint32_t mUnpackSkipImages = 0;
-  uint32_t mUnpackRowLength = 0;
-  uint32_t mUnpackSkipRows = 0;
-  uint32_t mUnpackSkipPixels = 0;
-  uint32_t mUnpackAlignment = 0;
-  GLenum mColorspaceConversion = 0;
-  bool mFlipY = false;
-  bool mPremultiplyAlpha = false;
-  bool mRequireFastPath = false;
-};
-
-using WebGLTexUnpackVariant =
-    Variant<UniquePtr<webgl::TexUnpackBytes>,
-            UniquePtr<webgl::TexUnpackSurface>,
-            UniquePtr<webgl::TexUnpackImage>, WebGLTexPboOffset>;
-
-using MaybeWebGLTexUnpackVariant = Maybe<WebGLTexUnpackVariant>;
-
 struct WebGLContextOptions {
   bool alpha = true;
   bool depth = true;
@@ -377,6 +362,8 @@ struct WebGLContextOptions {
   bool xrCompatible = false;
   dom::WebGLPowerPreference powerPreference =
       dom::WebGLPowerPreference::Default;
+  dom::PredefinedColorSpace colorSpace = dom::PredefinedColorSpace::Srgb;
+  bool ignoreColorSpace = true;  // Our legacy behavior.
   bool shouldResistFingerprinting = true;
   bool enableDebugRendererInfo = false;
 
@@ -389,10 +376,28 @@ struct WebGLContextOptions {
   }
 };
 
+namespace gfx {
+
+inline ColorSpace2 ToColorSpace2(const dom::PredefinedColorSpace cs) {
+  switch (cs) {
+    case dom::PredefinedColorSpace::Srgb:
+      return ColorSpace2::SRGB;
+    case dom::PredefinedColorSpace::Display_p3:
+      return ColorSpace2::DISPLAY_P3;
+    case dom::PredefinedColorSpace::EndGuard_:
+      break;
+  }
+  MOZ_CRASH("Exhaustive switch");
+}
+
+}  // namespace gfx
+
 // -
 
-template <typename T>
+template <typename _T>
 struct avec2 {
+  using T = _T;
+
   T x = T();
   T y = T();
 
@@ -418,10 +423,55 @@ struct avec2 {
 
   bool operator==(const avec2& rhs) const { return x == rhs.x && y == rhs.y; }
   bool operator!=(const avec2& rhs) const { return !(*this == rhs); }
+
+#define _(OP)                                 \
+  avec2 operator OP(const avec2& rhs) const { \
+    return {x OP rhs.x, y OP rhs.y};          \
+  }                                           \
+  avec2 operator OP(const T rhs) const { return {x OP rhs, y OP rhs}; }
+
+  _(+)
+  _(-)
+  _(*)
+  _(/)
+
+#undef _
+
+  avec2 Clamp(const avec2& min, const avec2& max) const {
+    return {mozilla::Clamp(x, min.x, max.x), mozilla::Clamp(y, min.y, max.y)};
+  }
+
+  // mozilla::Clamp doesn't work on floats, so be clear that this is a min+max
+  // helper.
+  avec2 ClampMinMax(const avec2& min, const avec2& max) const {
+    const auto ClampScalar = [](const T v, const T min, const T max) {
+      return std::max(min, std::min(v, max));
+    };
+    return {ClampScalar(x, min.x, max.x), ClampScalar(y, min.y, max.y)};
+  }
+
+  template <typename U>
+  U StaticCast() const {
+    return {static_cast<typename U::T>(x), static_cast<typename U::T>(y)};
+  }
 };
 
 template <typename T>
+avec2<T> MinExtents(const avec2<T>& a, const avec2<T>& b) {
+  return {std::min(a.x, b.x), std::min(a.y, b.y)};
+}
+
+template <typename T>
+avec2<T> MaxExtents(const avec2<T>& a, const avec2<T>& b) {
+  return {std::max(a.x, b.x), std::max(a.y, b.y)};
+}
+
+// -
+
+template <typename _T>
 struct avec3 {
+  using T = _T;
+
   T x = T();
   T y = T();
   T z = T();
@@ -449,10 +499,12 @@ struct avec3 {
   bool operator!=(const avec3& rhs) const { return !(*this == rhs); }
 };
 
-typedef avec2<int32_t> ivec2;
-typedef avec3<int32_t> ivec3;
-typedef avec2<uint32_t> uvec2;
-typedef avec3<uint32_t> uvec3;
+using ivec2 = avec2<int32_t>;
+using ivec3 = avec3<int32_t>;
+using uvec2 = avec2<uint32_t>;
+using uvec3 = avec3<uint32_t>;
+
+inline ivec2 AsVec(const gfx::IntSize& s) { return {s.width, s.height}; }
 
 // -
 
@@ -481,25 +533,16 @@ struct DriverUnpackInfo final {
   PackingInfo ToPacking() const { return {unpackFormat, unpackType}; }
 };
 
-struct PixelPackState final {
-  uint32_t alignment = 4;
-  uint32_t rowLength = 0;
-  uint32_t skipRows = 0;
-  uint32_t skipPixels = 0;
-};
+// -
 
-struct ReadPixelsDesc final {
-  ivec2 srcOffset;
-  uvec2 size;
-  PackingInfo pi = {LOCAL_GL_RGBA, LOCAL_GL_UNSIGNED_BYTE};
-  PixelPackState packState;
-};
-
-class ExtensionBits final {
+template <typename E>
+class EnumMask {
+ public:
   uint64_t mBits = 0;
 
+ private:
   struct BitRef final {
-    ExtensionBits& bits;
+    EnumMask& bits;
     const uint64_t mask;
 
     explicit operator bool() const { return bits.mBits & mask; }
@@ -514,15 +557,17 @@ class ExtensionBits final {
     }
   };
 
-  uint64_t Mask(const WebGLExtensionID i) const {
+  uint64_t Mask(const E i) const {
     return uint64_t{1} << static_cast<uint64_t>(i);
   }
 
  public:
-  BitRef operator[](const WebGLExtensionID i) { return {*this, Mask(i)}; }
-  bool operator[](const WebGLExtensionID i) const { return mBits & Mask(i); }
+  BitRef operator[](const E i) { return {*this, Mask(i)}; }
+  bool operator[](const E i) const { return mBits & Mask(i); }
 };
 
+class ExtensionBits : public EnumMask<WebGLExtensionID> {};
+
 // -
 
 enum class ContextLossReason : uint8_t {
@@ -550,6 +595,8 @@ struct InitContextDesc final {
   uint32_t principalKey = 0;
 };
 
+constexpr uint32_t kMaxTransformFeedbackSeparateAttribs = 4;
+
 struct Limits final {
   ExtensionBits supportedExtensions;
 
@@ -558,14 +605,13 @@ struct Limits final {
   uint32_t maxTex2dSize = 0;
   uint32_t maxTexCubeSize = 0;
   uint32_t maxVertexAttribs = 0;
-  std::array<uint32_t, 2> maxViewportDims = {};
+  uint32_t maxViewportDim = 0;
   std::array<float, 2> pointSizeRange = {{1, 1}};
   std::array<float, 2> lineWidthRange = {{1, 1}};
 
   // WebGL 2
   uint32_t maxTexArrayLayers = 0;
   uint32_t maxTex3dSize = 0;
-  uint32_t maxTransformFeedbackSeparateAttribs = 0;
   uint32_t maxUniformBufferBindings = 0;
   uint32_t uniformBufferOffsetAlignment = 0;
 
@@ -581,6 +627,7 @@ struct InitContextResult final {
   std::string error;
   WebGLContextOptions options;
   webgl::Limits limits;
+  EnumMask<layers::SurfaceDescriptor::Type> uploadableSdTypes;
 };
 
 // -
@@ -610,21 +657,24 @@ enum class LossStatus {
 
 struct CompileResult final {
   bool pending = true;
-  std::string log;
-  std::string translatedSource;
+  nsCString log;
+  nsCString translatedSource;
   bool success = false;
 };
 
 // -
 
-struct OpaqueFramebufferOptions {
+struct OpaqueFramebufferOptions final {
   bool depthStencil = true;
   bool antialias = true;
   uint32_t width = 0;
   uint32_t height = 0;
+};
 
-  OpaqueFramebufferOptions() = default;
-  OpaqueFramebufferOptions(const OpaqueFramebufferOptions&) = default;
+// -
+
+struct SwapChainOptions final {
+  bool bgra = false;
 };
 
 // -
@@ -668,7 +718,7 @@ struct LinkActiveInfo final {
 
 struct LinkResult final {
   bool pending = true;
-  std::string log;
+  nsCString log;
   bool success = false;
   LinkActiveInfo active;
   GLenum tfBufferMode = 0;
@@ -688,6 +738,20 @@ struct GetUniformData final {
   GLenum type = 0;
 };
 
+struct FrontBufferSnapshotIpc final {
+  uvec2 surfSize = {};
+  Maybe<mozilla::ipc::Shmem> shmem = {};
+};
+
+struct ReadPixelsResult {
+  gfx::IntRect subrect = {};
+  size_t byteStride = 0;
+};
+
+struct ReadPixelsResultIpc final : public ReadPixelsResult {
+  mozilla::ipc::Shmem shmem = {};
+};
+
 struct VertAttribPointerDesc final {
   bool intFunc = false;
   uint8_t channels = 4;
@@ -705,32 +769,15 @@ struct VertAttribPointerCalculated final {
 
 }  // namespace webgl
 
-// return value for the InitializeCanvasRenderer message
-struct ICRData {
-  gfx::IntSize size;
-  bool hasAlpha;
-  bool isPremultAlpha;
-};
-
 /**
  * Represents a block of memory that it may or may not own.  The
- * inner data type must be trivially copyable by memcpy.  A RawBuffer
- * may be backed by local memory or shared memory.
+ * inner data type must be trivially copyable by memcpy.
  */
-template <typename T = uint8_t, typename nonCV = std::remove_cv_t<T>,
-          std::enable_if_t<std::is_trivially_assignable<nonCV&, nonCV>::value,
-                           int> = 0>
-class RawBuffer {
-  // The SharedMemoryBasic that owns mData, if any.
-  RefPtr<mozilla::ipc::SharedMemoryBasic> mSmem;
-  // Pointer to the raw memory block
-  T* mData = nullptr;
-  // Length is the number of elements of size T in the array
-  size_t mLength = 0;
-  // true if we should delete[] the mData on destruction
-  bool mOwnsData = false;
-
-  friend struct mozilla::webgl::QueueParamTraits<RawBuffer<T>>;
+template <typename T = uint8_t>
+class RawBuffer final {
+  const T* mBegin = nullptr;
+  size_t mLen = 0;
+  UniqueBuffer mOwned;
 
  public:
   using ElementType = T;
@@ -738,129 +785,35 @@ class RawBuffer {
   /**
    * If aTakeData is true, RawBuffer will delete[] the memory when destroyed.
    */
-  RawBuffer(size_t len, T* data, bool aTakeData = false)
-      : mData(data), mLength(len), mOwnsData(aTakeData) {}
+  explicit RawBuffer(const Range<const T>& data, UniqueBuffer&& owned = {})
+      : mBegin(data.begin().get()),
+        mLen(data.length()),
+        mOwned(std::move(owned)) {}
 
-  RawBuffer(size_t len, RefPtr<mozilla::ipc::SharedMemoryBasic>& aSmem)
-      : mSmem(aSmem), mData(aSmem->memory()), mLength(len), mOwnsData(false) {
-    MOZ_ASSERT(mData && mLength);
-  }
+  explicit RawBuffer(const size_t len) : mLen(len) {}
 
-  ~RawBuffer() {
-    // If we have a SharedMemoryBasic then it must own mData.
-    MOZ_ASSERT((!mSmem) || (!mOwnsData));
-    if (mOwnsData) {
-      delete[] mData;
-    }
-    if (mSmem) {
-      mSmem->CloseHandle();
-    }
-  }
+  ~RawBuffer() = default;
 
-  auto Length() const { return mLength; }
+  Range<const T> Data() const { return {mBegin, mLen}; }
+  const auto& begin() const { return mBegin; }
+  const auto& size() const { return mLen; }
 
-  T* Data() { return mData; }
-  const T* Data() const { return mData; }
-
-  T& operator[](size_t idx) {
-    MOZ_ASSERT(mData && (idx < mLength));
-    return mData[idx];
-  }
-  const T& operator[](size_t idx) const {
-    MOZ_ASSERT(mData && (idx < mLength));
-    return mData[idx];
+  void Shrink(const size_t newLen) {
+    if (mLen <= newLen) return;
+    mLen = newLen;
   }
 
   RawBuffer() = default;
+
   RawBuffer(const RawBuffer&) = delete;
   RawBuffer& operator=(const RawBuffer&) = delete;
 
-  RawBuffer(RawBuffer&& o)
-      : mSmem(o.mSmem),
-        mData(o.mData),
-        mLength(o.mLength),
-        mOwnsData(o.mOwnsData) {
-    o.mSmem = nullptr;
-    o.mData = nullptr;
-    o.mLength = 0;
-    o.mOwnsData = false;
-  }
-
-  RawBuffer& operator=(RawBuffer&& o) {
-    mSmem = o.mSmem;
-    mData = o.mData;
-    mLength = o.mLength;
-    mOwnsData = o.mOwnsData;
-    o.mSmem = nullptr;
-    o.mData = nullptr;
-    o.mLength = 0;
-    o.mOwnsData = false;
-    return *this;
-  }
+  RawBuffer(RawBuffer&&) = default;
+  RawBuffer& operator=(RawBuffer&&) = default;
 };
 
 // -
 
-// clang-format off
-
-#define FOREACH_ID(X) \
-  X(FuncScopeIdError) \
-  X(compressedTexImage2D) \
-  X(compressedTexImage3D) \
-  X(compressedTexSubImage2D) \
-  X(compressedTexSubImage3D) \
-  X(copyTexSubImage2D) \
-  X(copyTexSubImage3D) \
-  X(drawArrays) \
-  X(drawArraysInstanced) \
-  X(drawElements) \
-  X(drawElementsInstanced) \
-  X(drawRangeElements) \
-  X(renderbufferStorage) \
-  X(renderbufferStorageMultisample) \
-  X(texImage2D) \
-  X(texImage3D) \
-  X(TexStorage2D) \
-  X(TexStorage3D) \
-  X(texSubImage2D) \
-  X(texSubImage3D) \
-  X(vertexAttrib1f) \
-  X(vertexAttrib1fv) \
-  X(vertexAttrib2f) \
-  X(vertexAttrib2fv) \
-  X(vertexAttrib3f) \
-  X(vertexAttrib3fv) \
-  X(vertexAttrib4f) \
-  X(vertexAttrib4fv) \
-  X(vertexAttribI4i) \
-  X(vertexAttribI4iv) \
-  X(vertexAttribI4ui) \
-  X(vertexAttribI4uiv) \
-  X(vertexAttribIPointer) \
-  X(vertexAttribPointer)
-
-// clang-format on
-
-enum class FuncScopeId {
-#define _(X) X,
-  FOREACH_ID(_)
-#undef _
-};
-
-static constexpr const char* const FUNCSCOPE_NAME_BY_ID[] = {
-#define _(X) #X,
-    FOREACH_ID(_)
-#undef _
-};
-
-#undef FOREACH_ID
-
-inline auto GetFuncScopeName(const FuncScopeId id) {
-  return FUNCSCOPE_NAME_BY_ID[static_cast<size_t>(id)];
-}
-
-// -
-
 template <typename C, typename K>
 inline auto MaybeFind(C& container, const K& key)
     -> decltype(&(container.find(key)->second)) {
@@ -903,12 +856,24 @@ inline GLenum ImageToTexTarget(const GLenum imageTarget) {
   }
 }
 
+inline bool IsTexTarget3D(const GLenum texTarget) {
+  switch (texTarget) {
+    case LOCAL_GL_TEXTURE_2D_ARRAY:
+    case LOCAL_GL_TEXTURE_3D:
+      return true;
+
+    default:
+      return false;
+  }
+}
+
 // -
 
 namespace dom {
 class Element;
 class ImageBitmap;
 class ImageData;
+class OffscreenCanvas;
 }  // namespace dom
 
 struct TexImageSource {
@@ -921,10 +886,122 @@ struct TexImageSource {
   const dom::ImageBitmap* mImageBitmap = nullptr;
   const dom::ImageData* mImageData = nullptr;
 
+  const dom::OffscreenCanvas* mOffscreenCanvas = nullptr;
+
   const dom::Element* mDomElem = nullptr;
   ErrorResult* mOut_error = nullptr;
 };
 
+namespace webgl {
+
+template <class DerivedT>
+struct DeriveNotEq {
+  bool operator!=(const DerivedT& rhs) const {
+    const auto self = reinterpret_cast<const DerivedT*>(this);
+    return !(*self == rhs);
+  }
+};
+
+struct PixelPackingState : public DeriveNotEq<PixelPackingState> {
+  uint32_t alignmentInTypeElems = 4;  // ALIGNMENT isn't naive byte alignment!
+  uint32_t rowLength = 0;
+  uint32_t imageHeight = 0;
+  uint32_t skipPixels = 0;
+  uint32_t skipRows = 0;
+  uint32_t skipImages = 0;
+
+  // C++20's default comparison operators can't come soon enough!
+  bool operator==(const PixelPackingState& rhs) const {
+    return alignmentInTypeElems == rhs.alignmentInTypeElems &&
+           rowLength == rhs.rowLength && imageHeight == rhs.imageHeight &&
+           skipPixels == rhs.skipPixels && skipRows == rhs.skipRows &&
+           skipImages == rhs.skipImages;
+  }
+
+  static void AssertDefaultUnpack(gl::GLContext& gl, const bool isWebgl2) {
+    PixelPackingState{}.AssertCurrentUnpack(gl, isWebgl2);
+  }
+
+  void ApplyUnpack(gl::GLContext&, bool isWebgl2,
+                   const uvec3& uploadSize) const;
+  bool AssertCurrentUnpack(gl::GLContext&, bool isWebgl2) const;
+};
+
+struct PixelUnpackStateWebgl final : public PixelPackingState {
+  GLenum colorspaceConversion =
+      dom::WebGLRenderingContext_Binding::BROWSER_DEFAULT_WEBGL;
+  bool flipY = false;
+  bool premultiplyAlpha = false;
+  bool requireFastPath = false;
+};
+
+struct ExplicitPixelPackingState final {
+  struct Metrics final {
+    uvec3 usedSize = {};
+    size_t bytesPerPixel = 0;
+
+    // (srcStrideAndRowOverride.x, otherwise ROW_LENGTH != 0, otherwise size.x)
+    // ...aligned to ALIGNMENT.
+    size_t bytesPerRowStride = 0;
+
+    // structuredSrcSize.y, otherwise IMAGE_HEIGHT*(SKIP_IMAGES+size.z)
+    size_t totalRows = 0;
+
+    // This ensures that no one else needs to do CheckedInt math.
+    size_t totalBytesUsed = 0;
+    size_t totalBytesStrided = 0;
+  };
+
+  // It's so important that these aren't modified once evaluated.
+  const PixelPackingState state;
+  const Metrics metrics;
+
+  static Result<ExplicitPixelPackingState, std::string> ForUseWith(
+      const PixelPackingState&, GLenum target, const uvec3& subrectSize,
+      const webgl::PackingInfo&, const Maybe<size_t> bytesPerRowStrideOverride);
+};
+
+struct ReadPixelsDesc final {
+  ivec2 srcOffset;
+  uvec2 size;
+  PackingInfo pi = {LOCAL_GL_RGBA, LOCAL_GL_UNSIGNED_BYTE};
+  PixelPackingState packState;
+};
+
+}  // namespace webgl
+
+namespace webgl {
+
+struct TexUnpackBlobDesc final {
+  GLenum imageTarget = LOCAL_GL_TEXTURE_2D;
+  uvec3 size;
+  gfxAlphaType srcAlphaType = gfxAlphaType::NonPremult;
+
+  Maybe<RawBuffer<>> cpuData;
+  Maybe<uint64_t> pboOffset;
+
+  Maybe<uvec2> structuredSrcSize;
+  RefPtr<layers::Image> image;
+  Maybe<layers::SurfaceDescriptor> sd;
+  RefPtr<gfx::DataSourceSurface> dataSurf;
+
+  webgl::PixelUnpackStateWebgl unpacking;
+  bool applyUnpackTransforms = true;
+
+  // -
+
+  auto ExplicitUnpacking(const webgl::PackingInfo& pi,
+                         const Maybe<size_t> bytesPerRowStrideOverride) const {
+    return ExplicitPixelPackingState::ForUseWith(this->unpacking,
+                                                 this->imageTarget, this->size,
+                                                 pi, bytesPerRowStrideOverride);
+  }
+
+  void Shrink(const webgl::PackingInfo&);
+};
+
+}  // namespace webgl
+
 // ---------------------------------------
 // MakeRange
 
@@ -940,12 +1017,7 @@ inline Range<const T> MakeRange(const dom::Sequence<T>& seq) {
 
 template <typename T>
 inline Range<const T> MakeRange(const RawBuffer<T>& from) {
-  return {from.Data(), from.Length()};
-}
-
-template <typename T>
-inline Range<T> MakeRange(RawBuffer<T>& from) {
-  return {from.Data(), from.Length()};
+  return from.Data();
 }
 
 // abv = ArrayBufferView
@@ -956,6 +1028,24 @@ inline auto MakeRangeAbv(const T& abv)
   return {abv.Data(), abv.Length()};
 }
 
+// -
+
+constexpr auto kUniversalAlignment = alignof(std::max_align_t);
+
+template <typename T>
+inline size_t AlignmentOffset(const size_t alignment, const T posOrPtr) {
+  MOZ_ASSERT(alignment);
+  const auto begin = reinterpret_cast<uintptr_t>(posOrPtr);
+  const auto wholeMultiples = (begin + (alignment - 1)) / alignment;
+  const auto aligned = wholeMultiples * alignment;
+  return aligned - begin;
+}
+
+template <typename T>
+inline size_t ByteSize(const Range<T>& range) {
+  return range.length() * sizeof(T);
+}
+
 Maybe<Range<const uint8_t>> GetRangeFromView(const dom::ArrayBufferView& view,
                                              GLuint elemOffset,
                                              GLuint elemCountOverride);
@@ -964,7 +1054,7 @@ Maybe<Range<const uint8_t>> GetRangeFromView(const dom::ArrayBufferView& view,
 
 template <typename T>
 RawBuffer<T> RawBufferView(const Range<T>& range) {
-  return {range.length(), range.begin().get()};
+  return RawBuffer<T>{range};
 }
 
 // -
@@ -989,6 +1079,27 @@ inline std::string ToString(const nsACString& text) {
   return {text.BeginReading(), text.Length()};
 }
 
+inline void Memcpy(const RangedPtr<uint8_t>& destBytes,
+                   const RangedPtr<const uint8_t>& srcBytes,
+                   const size_t byteSize) {
+  // Trigger range asserts
+  (void)(srcBytes + byteSize);
+  (void)(destBytes + byteSize);
+
+  memcpy(destBytes.get(), srcBytes.get(), byteSize);
+}
+
+// -
+
+namespace webgl {
+
+// In theory, this number can be unbounded based on the driver. However, no
+// driver appears to expose more than 8. We might as well stop there too, for
+// now.
+// (http://opengl.gpuinfo.org/gl_stats_caps_single.php?listreportsbycap=GL_MAX_COLOR_ATTACHMENTS)
+inline constexpr size_t kMaxDrawBuffers = 8;
+}  // namespace webgl
+
 }  // namespace mozilla
 
 #endif
