@@ -7,8 +7,8 @@
 #define mozilla_dom_HTMLSelectElement_h
 
 #include "mozilla/Attributes.h"
+#include "mozilla/dom/ConstraintValidation.h"
 #include "nsGenericHTMLElement.h"
-#include "nsIConstraintValidation.h"
 
 #include "mozilla/dom/BindingDeclarations.h"
 #include "mozilla/dom/UnionTypes.h"
@@ -34,7 +34,7 @@ class PresState;
 
 namespace dom {
 
-class HTMLFormSubmission;
+class FormData;
 class HTMLSelectElement;
 
 class MOZ_STACK_CLASS SafeOptionListMutation {
@@ -64,8 +64,8 @@ class MOZ_STACK_CLASS SafeOptionListMutation {
   /** Whether we should be notifying when we make various method calls on
       mSelect */
   const bool mNotify;
-  /** The selected index at mutation start. */
-  int32_t mInitialSelectedIndex;
+  /** The selected option at mutation start. */
+  RefPtr<HTMLOptionElement> mInitialSelectedOption;
   /** Option list must be recreated if more than one mutation is detected. */
   nsMutationGuard mGuard;
 };
@@ -73,8 +73,8 @@ class MOZ_STACK_CLASS SafeOptionListMutation {
 /**
  * Implementation of &lt;select&gt;
  */
-class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
-                                public nsIConstraintValidation {
+class HTMLSelectElement final : public nsGenericHTMLFormControlElementWithState,
+                                public ConstraintValidation {
  public:
   /**
    *  IS_SELECTED   whether to set the option(s) to true or false
@@ -98,7 +98,7 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
     NO_RESELECT = 1 << 4
   };
 
-  using nsIConstraintValidation::GetValidationMessage;
+  using ConstraintValidation::GetValidationMessage;
 
   explicit HTMLSelectElement(
       already_AddRefed<mozilla::dom::NodeInfo>&& aNodeInfo,
@@ -130,9 +130,6 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
   void SetDisabled(bool aVal, ErrorResult& aRv) {
     SetHTMLBoolAttr(nsGkAtoms::disabled, aVal, aRv);
   }
-  HTMLFormElement* GetForm() const {
-    return nsGenericHTMLFormElementWithState::GetForm();
-  }
   bool Multiple() const { return GetBoolAttr(nsGkAtoms::multiple); }
   void SetMultiple(bool aVal, ErrorResult& aRv) {
     SetHTMLBoolAttr(nsGkAtoms::multiple, aVal, aRv);
@@ -167,7 +164,7 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
   }
   void Add(const HTMLOptionElementOrHTMLOptGroupElement& aElement,
            const Nullable<HTMLElementOrLong>& aBefore, ErrorResult& aRv);
-  void Remove(int32_t aIndex);
+  void Remove(int32_t aIndex) const;
   void IndexedSetter(uint32_t aIndex, HTMLOptionElement* aOption,
                      ErrorResult& aRv) {
     mOptions->IndexedSetter(aIndex, aOption, aRv);
@@ -179,7 +176,7 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
 
   int32_t SelectedIndex() const { return mSelectedIndex; }
   void SetSelectedIndex(int32_t aIdx) { SetSelectedIndexInternal(aIdx, true); }
-  void GetValue(DOMString& aValue);
+  void GetValue(DOMString& aValue) const;
   void SetValue(const nsAString& aValue);
 
   // Override SetCustomValidity so we update our state properly when it's called
@@ -203,12 +200,16 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
                                  bool aNotify, ErrorResult& aRv) override;
   virtual void RemoveChildNode(nsIContent* aKid, bool aNotify) override;
 
+  // nsGenericHTMLElement
+  virtual bool IsDisabledForEvents(WidgetEvent* aEvent) override;
+
+  // nsGenericHTMLFormElement
+  void SaveState() override;
+  bool RestoreState(PresState* aState) override;
+
   // Overriden nsIFormControl methods
   NS_IMETHOD Reset() override;
-  NS_IMETHOD SubmitNamesValues(HTMLFormSubmission* aFormSubmission) override;
-  NS_IMETHOD SaveState() override;
-  virtual bool RestoreState(PresState* aState) override;
-  virtual bool IsDisabledForEvents(WidgetEvent* aEvent) override;
+  NS_IMETHOD SubmitNamesValues(FormData* aFormData) override;
 
   virtual void FieldSetDisabledChanged(bool aNotify) override;
 
@@ -292,12 +293,12 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
 
   virtual nsresult Clone(dom::NodeInfo*, nsINode** aResult) const override;
 
-  NS_DECL_CYCLE_COLLECTION_CLASS_INHERITED(HTMLSelectElement,
-                                           nsGenericHTMLFormElementWithState)
+  NS_DECL_CYCLE_COLLECTION_CLASS_INHERITED(
+      HTMLSelectElement, nsGenericHTMLFormControlElementWithState)
 
   HTMLOptionsCollection* GetOptions() { return mOptions; }
 
-  // nsIConstraintValidation
+  // ConstraintValidation
   nsresult GetValidationMessage(nsAString& aValidationMessage,
                                 ValidityStateType aType) override;
 
@@ -321,8 +322,8 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
    */
   bool IsCombobox() const { return !Multiple() && Size() <= 1; }
 
-  bool OpenInParentProcess();
-  void SetOpenInParentProcess(bool aVal);
+  bool OpenInParentProcess() const { return mIsOpenInParentProcess; }
+  void SetOpenInParentProcess(bool aVal) { mIsOpenInParentProcess = aVal; }
 
   void GetPreviewValue(nsAString& aValue) { aValue = mPreviewValue; }
   void SetPreviewValue(const nsAString& aValue);
@@ -338,7 +339,7 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
    * @param aIndex the index
    * @return whether the option at the index is selected
    */
-  bool IsOptionSelectedByIndex(int32_t aIndex);
+  bool IsOptionSelectedByIndex(int32_t aIndex) const;
   /**
    * Starting with (and including) aStartIndex, find the first selected index
    * and set mSelectedIndex to it.
@@ -436,8 +437,8 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
   nsISelectControlFrame* GetSelectFrame();
 
   /**
-   * Helper method for dispatching ContentReset notifications to list
-   * and combo box frames.
+   * Helper method for dispatching ContentReset notifications to list box
+   * frames.
    */
   void DispatchContentReset();
 
@@ -485,33 +486,36 @@ class HTMLSelectElement final : public nsGenericHTMLFormElementWithState,
   nsContentUtils::AutocompleteAttrState mAutocompleteAttrState;
   nsContentUtils::AutocompleteAttrState mAutocompleteInfoState;
   /** false if the parser is in the middle of adding children. */
-  bool mIsDoneAddingChildren;
+  bool mIsDoneAddingChildren : 1;
   /** true if our disabled state has changed from the default **/
-  bool mDisabledChanged;
+  bool mDisabledChanged : 1;
   /** true if child nodes are being added or removed.
    *  Used by SafeOptionListMutation.
    */
-  bool mMutating;
+  bool mMutating : 1;
   /**
    * True if DoneAddingChildren will get called but shouldn't restore state.
    */
-  bool mInhibitStateRestoration;
+  bool mInhibitStateRestoration : 1;
   /**
    * True if the selection has changed since the element's creation.
    */
-  bool mSelectionHasChanged;
+  bool mSelectionHasChanged : 1;
   /**
    * True if the default selected option has been set.
    */
-  bool mDefaultSelectionSet;
+  bool mDefaultSelectionSet : 1;
   /**
    * True if :-moz-ui-invalid can be shown.
    */
-  bool mCanShowInvalidUI;
+  bool mCanShowInvalidUI : 1;
   /**
    * True if :-moz-ui-valid can be shown.
    */
-  bool mCanShowValidUI;
+  bool mCanShowValidUI : 1;
+
+  /** True if we're open in the parent process */
+  bool mIsOpenInParentProcess : 1;
 
   /** The number of non-options as children of the select */
   uint32_t mNonOptionChildren;
