@@ -16,6 +16,7 @@
 #  include "SeekTarget.h"
 #  include "mozilla/Atomics.h"
 #  include "mozilla/Maybe.h"
+#  include "mozilla/MozPromise.h"
 #  include "mozilla/Mutex.h"
 #  include "mozilla/StateMirroring.h"
 #  include "mozilla/StaticPrefs_media.h"
@@ -77,7 +78,6 @@ class MediaFormatReader final
   typedef MozPromise<bool, MediaResult, IsExclusive> NotifyDataArrivedPromise;
 
  public:
-  MOZ_DECLARE_THREADSAFEWEAKREFERENCE_TYPENAME(MediaFormatReader)
   MOZ_DECLARE_REFCOUNTED_TYPENAME(MediaFormatReader)
 
   using TrackSet = EnumSet<TrackInfo::TrackType>;
@@ -133,6 +133,10 @@ class MediaFormatReader final
   // mBuffered should be recalculated and updated accordingly.
   void NotifyDataArrived();
 
+  // Update ID for the external playback engine. Currently it's only used on
+  // Windows when the media engine playback is enabled.
+  void UpdateMediaEngineId(uint64_t aMediaEngineId);
+
  protected:
   // Recomputes mBuffered.
   void UpdateBuffered();
@@ -183,9 +187,11 @@ class MediaFormatReader final
 
   RefPtr<SetCDMPromise> SetCDMProxy(CDMProxy* aProxy);
 
-  // Returns a MediaDebugInfo structure
-  // Used for debugging purposes.
-  void GetDebugInfo(dom::MediaFormatReaderDebugInfo& aInfo);
+  // Requests that the MediaFormatReader populates aInfo with debug information.
+  // This may be done asynchronously, and aInfo should *not* be accessed by the
+  // caller until the returned promise is resolved or rejected.
+  RefPtr<GenericPromise> RequestDebugInfo(
+      dom::MediaFormatReaderDebugInfo& aInfo);
 
   // Switch the video decoder to NullDecoderModule. It might takes effective
   // since a few samples later depends on how much demuxed samples are already
@@ -359,8 +365,11 @@ class MediaFormatReader final
           mFlushing(false),
           mFlushed(true),
           mDrainState(DrainState::None),
-          mNumOfConsecutiveError(0),
-          mMaxConsecutiveError(aNumOfMaxError),
+          mNumOfConsecutiveDecodingError(0),
+          mMaxConsecutiveDecodingError(aNumOfMaxError),
+          mNumOfConsecutiveRDDOrGPUCrashes(0),
+          mMaxConsecutiveRDDOrGPUCrashes(
+              StaticPrefs::media_rdd_process_max_crashes()),
           mFirstFrameTime(Some(media::TimeUnit::Zero())),
           mNumSamplesInput(0),
           mNumSamplesOutput(0),
@@ -392,7 +401,7 @@ class MediaFormatReader final
     // as those can be read outside the TaskQueue.
     // They are only written on the TaskQueue however, as such mMutex doesn't
     // need to be held when those members are read on the TaskQueue.
-    Mutex mMutex;
+    Mutex mMutex MOZ_UNANNOTATED;
     // The platform decoder.
     RefPtr<MediaDataDecoder> mDecoder;
     nsCString mDescription;
@@ -448,8 +457,14 @@ class MediaFormatReader final
       mDrainState = DrainState::DrainRequested;
     }
 
-    uint32_t mNumOfConsecutiveError;
-    uint32_t mMaxConsecutiveError;
+    // Track decoding error and fail when we hit the limit.
+    uint32_t mNumOfConsecutiveDecodingError;
+    uint32_t mMaxConsecutiveDecodingError;
+
+    // Track RDD process crashes and fail when we hit the limit.
+    uint32_t mNumOfConsecutiveRDDOrGPUCrashes;
+    uint32_t mMaxConsecutiveRDDOrGPUCrashes;
+
     // Set when we haven't yet decoded the first frame.
     // Cleared once the first frame has been decoded.
     // This is used to determine, upon error, if we should try again to decode
@@ -464,12 +479,19 @@ class MediaFormatReader final
       if (mError.ref() == NS_ERROR_DOM_MEDIA_DECODE_ERR) {
         // Allow decode errors to be non-fatal, but give up
         // if we have too many, or if warnings should be treated as errors.
-        return mNumOfConsecutiveError > mMaxConsecutiveError ||
+        return mNumOfConsecutiveDecodingError > mMaxConsecutiveDecodingError ||
                StaticPrefs::media_playback_warnings_as_errors();
       } else if (mError.ref() == NS_ERROR_DOM_MEDIA_NEED_NEW_DECODER) {
         // If the caller asked for a new decoder we shouldn't treat
         // it as fatal.
         return false;
+      } else if (mError.ref() ==
+                 NS_ERROR_DOM_MEDIA_REMOTE_DECODER_CRASHED_RDD_OR_GPU_ERR) {
+        // Allow RDD crashes to be non-fatal, but give up
+        // if we have too many, or if warnings should be treated as errors.
+        return mNumOfConsecutiveRDDOrGPUCrashes >
+                   mMaxConsecutiveRDDOrGPUCrashes ||
+               StaticPrefs::media_playback_warnings_as_errors();
       } else {
         // All other error types are fatal
         return true;
@@ -586,6 +608,9 @@ class MediaFormatReader final
     // Use NullDecoderModule or not.
     bool mIsNullDecode;
     bool mHardwareDecodingDisabled;
+    // Whether we have reported hardware decoding support for video. Used only
+    // on reader's task queue,
+    bool mHasReportedVideoHardwareSupportTelemtry = false;
 
     class {
      public:
@@ -807,6 +832,11 @@ class MediaFormatReader final
   MozPromiseHolder<SetCDMPromise> mSetCDMPromise;
   TrackSet mSetCDMForTracks{};
   bool IsDecoderWaitingForCDM(TrackType aTrack);
+
+  void GetDebugInfo(dom::MediaFormatReaderDebugInfo& aInfo);
+
+  // Only be used on Windows when the media engine playback is enabled.
+  Maybe<uint64_t> mMediaEngineId;
 };
 
 }  // namespace mozilla
