@@ -16,38 +16,32 @@
 #include "mozilla/StaticPrefs_media.h"
 #include "mozilla/StaticPrefs_dom.h"
 #include "nsPrintfCString.h"
+#include "Tracing.h"
 
 namespace mozilla {
 
-extern LazyLogModule gMediaDecoderLog;
-#define SINK_LOG(msg, ...)                   \
-  MOZ_LOG(gMediaDecoderLog, LogLevel::Debug, \
+mozilla::LazyLogModule gAudioSinkLog("AudioSink");
+#define SINK_LOG(msg, ...)                \
+  MOZ_LOG(gAudioSinkLog, LogLevel::Debug, \
           ("AudioSink=%p " msg, this, ##__VA_ARGS__))
-#define SINK_LOG_V(msg, ...)                   \
-  MOZ_LOG(gMediaDecoderLog, LogLevel::Verbose, \
+#define SINK_LOG_V(msg, ...)                \
+  MOZ_LOG(gAudioSinkLog, LogLevel::Verbose, \
           ("AudioSink=%p " msg, this, ##__VA_ARGS__))
 
 // The amount of audio frames that is used to fuzz rounding errors.
 static const int64_t AUDIO_FUZZ_FRAMES = 1;
 
-// Amount of audio frames we will be processing ahead of use
-static const int32_t LOW_AUDIO_USECS = 300000;
-
 using media::TimeUnit;
 
 AudioSink::AudioSink(AbstractThread* aThread,
-                     MediaQueue<AudioData>& aAudioQueue,
-                     const TimeUnit& aStartTime, const AudioInfo& aInfo,
+                     MediaQueue<AudioData>& aAudioQueue, const AudioInfo& aInfo,
                      AudioDeviceInfo* aAudioDevice)
-    : mStartTime(aStartTime),
-      mInfo(aInfo),
+    : mInfo(aInfo),
       mAudioDevice(aAudioDevice),
       mPlaying(true),
-      mMonitor("AudioSink"),
       mWritten(0),
       mErrored(false),
       mOwnerThread(aThread),
-      mProcessedQueueLength(0),
       mFramesParsed(0),
       mOutputRate(DecideAudioPlaybackSampleRate(aInfo)),
       mOutputChannels(DecideAudioPlaybackChannels(aInfo)),
@@ -55,29 +49,119 @@ AudioSink::AudioSink(AbstractThread* aThread,
           mOutputRate,
           StaticPrefs::dom_media_silence_duration_for_audibility()),
       mIsAudioDataAudible(false),
-      mAudioQueue(aAudioQueue) {}
+      mProcessedQueueFinished(false),
+      mAudioQueue(aAudioQueue),
+      mProcessedQueueThresholdMS(
+          StaticPrefs::media_audio_audiosink_threshold_ms()) {
+  // Not much to initialize here if there's no audio.
+  if (!aInfo.IsValid()) {
+    mProcessedSPSCQueue = MakeUnique<SPSCQueue<AudioDataValue>>(0);
+    return;
+  }
+  // Twice the limit that trigger a refill.
+  double capacitySeconds = mProcessedQueueThresholdMS / 1000.f * 2;
+  // Clamp to correct boundaries, and align on the channel count
+  int elementCount = static_cast<int>(
+      std::clamp(capacitySeconds * mOutputChannels * mOutputRate, 0.,
+                 std::numeric_limits<int>::max() - 1.));
+  elementCount -= elementCount % mOutputChannels;
+  mProcessedSPSCQueue = MakeUnique<SPSCQueue<AudioDataValue>>(elementCount);
+  SINK_LOG("Ringbuffer has space for %u elements (%lf seconds)",
+           mProcessedSPSCQueue->Capacity(),
+           static_cast<float>(elementCount) / mOutputChannels / mOutputRate);
+  // Determine if the data is likely to be audible when the stream will be
+  // ready, if possible.
+  RefPtr<AudioData> frontPacket = mAudioQueue.PeekFront();
+  if (frontPacket) {
+    mAudibilityMonitor.ProcessInterleaved(frontPacket->Data(),
+                                          frontPacket->mChannels);
+    mIsAudioDataAudible = mAudibilityMonitor.RecentlyAudible();
+    SINK_LOG("New AudioSink -- audio is likely to be %s",
+             mIsAudioDataAudible ? "audible" : "inaudible");
+  } else {
+    // If no packets are available, consider the audio audible.
+    mIsAudioDataAudible = true;
+    SINK_LOG(
+        "New AudioSink -- no audio packet avaialble, considering the stream "
+        "audible");
+  }
+}
 
-AudioSink::~AudioSink() = default;
+AudioSink::~AudioSink() {
+  // Generally instances of AudioSink should be properly Shutdown manually.
+  // The only way deleting an AudioSink without shutdown an happen is if the
+  // dispatch back to the MDSM thread after initializing it asynchronously
+  // fails. When that's the case, the stream has been initialized but not
+  // started. Manually shutdown the AudioStream in this case.
+  if (mAudioStream) {
+    mAudioStream->Shutdown();
+  }
+}
 
-Result<already_AddRefed<MediaSink::EndedPromise>, nsresult> AudioSink::Start(
-    const PlaybackParams& aParams) {
+nsresult AudioSink::InitializeAudioStream(
+    const PlaybackParams& aParams,
+    AudioSink::InitializationType aInitializationType) {
+  if (aInitializationType == AudioSink::InitializationType::UNMUTING) {
+    // Consider the stream to be audible immediately, before initialization
+    // finishes when unmuting, in case initialization takes some time and it
+    // looked audible when the AudioSink was created.
+    mAudibleEvent.Notify(mIsAudioDataAudible);
+    SINK_LOG("InitializeAudioStream (Unmuting) notifying that audio is %s",
+             mIsAudioDataAudible ? "audible" : "inaudible");
+  } else {
+    // If not unmuting, the audibility event will be dispatched as usual,
+    // inspecting the audio content as it's being played and signaling the
+    // audibility event when a different in state is detected.
+    SINK_LOG("InitializeAudioStream (initial)");
+    mIsAudioDataAudible = false;
+  }
+
+  // When AudioQueue is empty, there is no way to know the channel layout of
+  // the coming audio data, so we use the predefined channel map instead.
+  AudioConfig::ChannelLayout::ChannelMap channelMap =
+      AudioConfig::ChannelLayout(mOutputChannels).Map();
+  // The layout map used here is already processed by mConverter with
+  // mOutputChannels into SMPTE format, so there is no need to worry if
+  // StaticPrefs::accessibility_monoaudio_enable() or
+  // StaticPrefs::media_forcestereo_enabled() is applied.
+  MOZ_ASSERT(!mAudioStream);
+  mAudioStream =
+      new AudioStream(*this, mOutputRate, mOutputChannels, channelMap);
+  nsresult rv = mAudioStream->Init(mAudioDevice);
+  if (NS_FAILED(rv)) {
+    mAudioStream->Shutdown();
+    mAudioStream = nullptr;
+    return rv;
+  }
+
+  // Set playback params before calling Start() so they can take effect
+  // as soon as the 1st DataCallback of the AudioStream fires.
+  mAudioStream->SetVolume(aParams.mVolume);
+  mAudioStream->SetPlaybackRate(aParams.mPlaybackRate);
+  mAudioStream->SetPreservesPitch(aParams.mPreservesPitch);
+
+  return NS_OK;
+}
+
+nsresult AudioSink::Start(
+    const media::TimeUnit& aStartTime,
+    MozPromiseHolder<MediaSink::EndedPromise>& aEndedPromise) {
   MOZ_ASSERT(mOwnerThread->IsCurrentThreadIn());
 
   mAudioQueueListener = mAudioQueue.PushEvent().Connect(
       mOwnerThread, this, &AudioSink::OnAudioPushed);
   mAudioQueueFinishListener = mAudioQueue.FinishEvent().Connect(
       mOwnerThread, this, &AudioSink::NotifyAudioNeeded);
-  mProcessedQueueListener = mProcessedQueue.PopFrontEvent().Connect(
-      mOwnerThread, this, &AudioSink::OnAudioPopped);
+  mProcessedQueueListener =
+      mAudioPopped.Connect(mOwnerThread, this, &AudioSink::OnAudioPopped);
+
+  mStartTime = aStartTime;
 
   // To ensure at least one audio packet will be popped from AudioQueue and
   // ready to be played.
   NotifyAudioNeeded();
-  nsresult rv = InitializeAudioStream(aParams);
-  if (NS_FAILED(rv)) {
-    return Err(rv);
-  }
-  return mAudioStream->Start();
+
+  return mAudioStream->Start(aEndedPromise);
 }
 
 TimeUnit AudioSink::GetPosition() {
@@ -103,58 +187,94 @@ TimeUnit AudioSink::GetPosition() {
 bool AudioSink::HasUnplayedFrames() {
   // Experimentation suggests that GetPositionInFrames() is zero-indexed,
   // so we need to add 1 here before comparing it to mWritten.
-  int64_t total;
-  {
-    MonitorAutoLock mon(mMonitor);
-    total = mWritten + (mCursor.get() ? mCursor->Available() : 0);
-  }
-  return mProcessedQueue.GetSize() ||
-         (mAudioStream && mAudioStream->GetPositionInFrames() + 1 < total);
+  return mProcessedSPSCQueue->AvailableRead() ||
+         (mAudioStream && mAudioStream->GetPositionInFrames() + 1 < mWritten);
 }
 
-void AudioSink::Shutdown() {
+TimeUnit AudioSink::UnplayedDuration() const {
+  return TimeUnit::FromMicroseconds(AudioQueuedInRingBufferMS());
+}
+
+void AudioSink::ReenqueueUnplayedAudioDataIfNeeded() {
+  // This is OK: the AudioStream has been shut down. Shutdown guarantees that
+  // the audio callback thread won't call back again.
+  mProcessedSPSCQueue->ResetThreadIds();
+
+  // construct an AudioData
+  int sampleCount = mProcessedSPSCQueue->AvailableRead();
+
+  if (!sampleCount) {
+    return;
+  }
+
+  uint32_t channelCount;
+  uint32_t rate;
+  if (mConverter) {
+    channelCount = mConverter->OutputConfig().Channels();
+    rate = mConverter->OutputConfig().Rate();
+  } else {
+    channelCount = mOutputChannels;
+    rate = mOutputRate;
+  }
+
+  uint32_t frameCount = sampleCount / channelCount;
+  auto duration = FramesToTimeUnit(frameCount, rate);
+  if (!duration.IsValid()) {
+    NS_WARNING("Int overflow in AudioSink");
+    mErrored = true;
+    return;
+  }
+
+  AlignedAudioBuffer queuedAudio(sampleCount);
+  DebugOnly<int> samplesRead =
+      mProcessedSPSCQueue->Dequeue(queuedAudio.Data(), sampleCount);
+  MOZ_ASSERT(samplesRead == sampleCount);
+
+  // Extrapolate mOffset, mTime from the front of the queue
+  // We can't really find a good value for `mOffset`, so we take what we have
+  // at the front of the queue.
+  // For `mTime`, assume there hasn't been a discontinuity recently.
+  RefPtr<AudioData> frontPacket = mAudioQueue.PeekFront();
+  uint32_t offset;
+  TimeUnit time;
+  if (!frontPacket) {
+    // We do our best here, but it's not going to be perfect.
+    offset = 0;
+    time = std::max(GetPosition() - duration, TimeUnit::Zero());
+  } else {
+    offset = frontPacket->mOffset;
+    time = frontPacket->mTime - duration;
+  }
+  RefPtr<AudioData> data =
+      new AudioData(offset, time, std::move(queuedAudio), channelCount, rate);
+  MOZ_DIAGNOSTIC_ASSERT(duration == data->mDuration, "must be equal");
+
+  SINK_LOG(
+      "Muting: Pushing back %u frames (%lfms) from the ring buffer back into "
+      "the audio queue",
+      frameCount, static_cast<float>(frameCount) / rate);
+
+  mAudioQueue.PushFront(data);
+}
+
+Maybe<MozPromiseHolder<MediaSink::EndedPromise>> AudioSink::Shutdown(
+    ShutdownCause aShutdownCause) {
   MOZ_ASSERT(mOwnerThread->IsCurrentThreadIn());
 
-  mAudioQueueListener.Disconnect();
-  mAudioQueueFinishListener.Disconnect();
-  mProcessedQueueListener.Disconnect();
+  mAudioQueueListener.DisconnectIfExists();
+  mAudioQueueFinishListener.DisconnectIfExists();
+  mProcessedQueueListener.DisconnectIfExists();
+
+  Maybe<MozPromiseHolder<MediaSink::EndedPromise>> rv;
 
   if (mAudioStream) {
-    mAudioStream->Shutdown();
+    rv = mAudioStream->Shutdown(aShutdownCause);
     mAudioStream = nullptr;
+    ReenqueueUnplayedAudioDataIfNeeded();
   }
-  // Shutdown audio sink doesn't mean the playback is going to stop, so if we
-  // simply discard these data, then we will no longer be able to play them.
-  // Eg. we change to sink to capture-based sink that will need to continue play
-  // remaining data from the audio queue.
-  {
-    MonitorAutoLock mon(mMonitor);
-    while (mProcessedQueue.GetSize() > 0) {
-      RefPtr<AudioData> audio = mProcessedQueue.PopBack();
-      if (audio == mCurrentData) {
-        break;
-      }
-      mAudioQueue.PushFront(audio);
-    }
-    if (mCurrentData) {
-      uint32_t unplayedFrames = mCursor->Available();
-      // If we've consumed some partial content from the first audio data, then
-      // we have to adjust its data offset and frames number in order not to
-      // play the same content again.
-      if (unplayedFrames > 0 && unplayedFrames < mCurrentData->Frames()) {
-        const uint32_t orginalFrames = mCurrentData->Frames();
-        const uint32_t offsetFrames = mCurrentData->Frames() - unplayedFrames;
-        Unused << mCurrentData->SetTrimWindow(
-            {mCurrentData->mTime + FramesToTimeUnit(offsetFrames, mOutputRate),
-             mCurrentData->GetEndTime()});
-        SINK_LOG_V("After adjustment, audio frame from %u to %u", orginalFrames,
-                   mCurrentData->Frames());
-      }
-      mAudioQueue.PushFront(mCurrentData);
-    }
-    MOZ_ASSERT(mProcessedQueue.GetSize() == 0);
-  }
-  mProcessedQueue.Finish();
+  mProcessedQueueFinished = true;
+
+  return rv;
 }
 
 void AudioSink::SetVolume(double aVolume) {
@@ -197,40 +317,8 @@ void AudioSink::SetPlaying(bool aPlaying) {
   mPlaying = aPlaying;
 }
 
-nsresult AudioSink::InitializeAudioStream(const PlaybackParams& aParams) {
-  mAudioStream = new AudioStream(*this);
-  // When AudioQueue is empty, there is no way to know the channel layout of
-  // the coming audio data, so we use the predefined channel map instead.
-  AudioConfig::ChannelLayout::ChannelMap channelMap =
-      mConverter ? mConverter->OutputConfig().Layout().Map()
-                 : AudioConfig::ChannelLayout(mOutputChannels).Map();
-  // The layout map used here is already processed by mConverter with
-  // mOutputChannels into SMPTE format, so there is no need to worry if
-  // StaticPrefs::accessibility_monoaudio_enable() or
-  // StaticPrefs::media_forcestereo_enabled() is applied.
-  nsresult rv = mAudioStream->Init(mOutputChannels, channelMap, mOutputRate,
-                                   mAudioDevice);
-  if (NS_FAILED(rv)) {
-    mAudioStream->Shutdown();
-    mAudioStream = nullptr;
-    return rv;
-  }
-
-  // Set playback params before calling Start() so they can take effect
-  // as soon as the 1st DataCallback of the AudioStream fires.
-  mAudioStream->SetVolume(aParams.mVolume);
-  mAudioStream->SetPlaybackRate(aParams.mPlaybackRate);
-  mAudioStream->SetPreservesPitch(aParams.mPreservesPitch);
-  return NS_OK;
-}
-
 TimeUnit AudioSink::GetEndTime() const {
-  int64_t written;
-  {
-    MonitorAutoLock mon(mMonitor);
-    written = mWritten;
-  }
-  TimeUnit played = FramesToTimeUnit(written, mOutputRate) + mStartTime;
+  TimeUnit played = FramesToTimeUnit(mWritten, mOutputRate) + mStartTime;
   if (!played.IsValid()) {
     NS_WARNING("Int overflow calculating audio end time");
     return TimeUnit::Zero();
@@ -240,109 +328,68 @@ TimeUnit AudioSink::GetEndTime() const {
   return std::min(mLastEndTime, played);
 }
 
-UniquePtr<AudioStream::Chunk> AudioSink::PopFrames(uint32_t aFrames) {
-  class Chunk : public AudioStream::Chunk {
-   public:
-    Chunk(AudioData* aBuffer, uint32_t aFrames, AudioDataValue* aData)
-        : mBuffer(aBuffer), mFrames(aFrames), mData(aData) {}
-    Chunk() : mFrames(0), mData(nullptr) {}
-    const AudioDataValue* Data() const override { return mData; }
-    uint32_t Frames() const override { return mFrames; }
-    uint32_t Channels() const override {
-      return mBuffer ? mBuffer->mChannels : 0;
-    }
-    uint32_t Rate() const override { return mBuffer ? mBuffer->mRate : 0; }
-    AudioDataValue* GetWritable() const override { return mData; }
-
-   private:
-    const RefPtr<AudioData> mBuffer;
-    const uint32_t mFrames;
-    AudioDataValue* const mData;
-  };
-
-  bool needPopping = false;
-  if (!mCurrentData) {
-    // No data in the queue. Return an empty chunk.
-    if (!mProcessedQueue.GetSize()) {
-      return MakeUnique<Chunk>();
-    }
-
-    // We need to update our values prior popping the processed queue in
-    // order to prevent the pop event to fire too early (prior
-    // mProcessedQueueLength being updated) or prevent HasUnplayedFrames
-    // to incorrectly return true during the time interval betweeen the
-    // when mProcessedQueue is read and mWritten is updated.
-    needPopping = true;
-    {
-      MonitorAutoLock mon(mMonitor);
-      mCurrentData = mProcessedQueue.PeekFront();
-      mCursor = MakeUnique<AudioBufferCursor>(mCurrentData->Data(),
-                                              mCurrentData->mChannels,
-                                              mCurrentData->Frames());
-    }
-    MOZ_ASSERT(mCurrentData->Frames() > 0);
-    mProcessedQueueLength -=
-        FramesToUsecs(mCurrentData->Frames(), mOutputRate).value();
+uint32_t AudioSink::PopFrames(AudioDataValue* aBuffer, uint32_t aFrames,
+                              bool aAudioThreadChanged) {
+  // This is safe, because we have the guarantee, by the OS, that audio
+  // callbacks are never called concurrently. Audio thread changes can only
+  // happen when not using cubeb remoting, and often when changing audio device
+  // at the system level.
+  if (aAudioThreadChanged) {
+    mProcessedSPSCQueue->ResetThreadIds();
   }
 
-  auto framesToPop = std::min(aFrames, mCursor->Available());
+  TRACE_COMMENT("AudioSink::PopFrames", "%u frames (ringbuffer: %u/%u)",
+                aFrames, SampleToFrame(mProcessedSPSCQueue->AvailableRead()),
+                SampleToFrame(mProcessedSPSCQueue->Capacity()));
 
-  SINK_LOG_V("playing audio at time=%" PRId64 " offset=%u length=%u",
-             mCurrentData->mTime.ToMicroseconds(),
-             mCurrentData->Frames() - mCursor->Available(), framesToPop);
-
-#ifdef MOZ_GECKO_PROFILER
-  mOwnerThread->Dispatch(NS_NewRunnableFunction(
-      "AudioSink:AddMarker",
-      [startTime = mCurrentData->mTime.ToMicroseconds(),
-       endTime = mCurrentData->GetEndTime().ToMicroseconds()] {
-        PROFILER_MARKER("PlayAudio", MEDIA_PLAYBACK, {}, MediaSampleMarker,
-                        startTime, endTime);
-      }));
-#endif  // MOZ_GECKO_PROFILER
-
-  UniquePtr<AudioStream::Chunk> chunk =
-      MakeUnique<Chunk>(mCurrentData, framesToPop, mCursor->Ptr());
-
-  {
-    MonitorAutoLock mon(mMonitor);
-    mWritten += framesToPop;
-    mCursor->Advance(framesToPop);
-    // All frames are popped. Reset mCurrentData so we can pop new elements from
-    // the audio queue in next calls to PopFrames().
-    if (!mCursor->Available()) {
-      mCurrentData = nullptr;
+  const int samplesToPop = static_cast<int>(aFrames * mOutputChannels);
+  const int samplesRead = mProcessedSPSCQueue->Dequeue(aBuffer, samplesToPop);
+  MOZ_ASSERT(samplesRead % mOutputChannels == 0);
+  mWritten += SampleToFrame(samplesRead);
+  if (samplesRead != samplesToPop) {
+    if (Ended()) {
+      SINK_LOG("Last PopFrames -- Source ended.");
+    } else {
+      NS_WARNING("Underrun when popping samples from audiosink ring buffer.");
+      TRACE_COMMENT("AudioSink::PopFrames", "Underrun %u frames missing",
+                    SampleToFrame(samplesToPop - samplesRead));
     }
+    // silence the rest
+    PodZero(aBuffer + samplesRead, samplesToPop - samplesRead);
   }
 
-  if (needPopping) {
-    // We can now safely pop the audio packet from the processed queue.
-    // This will fire the popped event, triggering a call to NotifyAudioNeeded.
-    RefPtr<AudioData> releaseMe = mProcessedQueue.PopFront();
-    CheckIsAudible(releaseMe);
-  }
+  mAudioPopped.Notify();
 
-  return chunk;
+  SINK_LOG_V("Popping %u frames. Remaining in ringbuffer %u / %u\n", aFrames,
+             SampleToFrame(mProcessedSPSCQueue->AvailableRead()),
+             SampleToFrame(mProcessedSPSCQueue->Capacity()));
+
+  // Don't consider the silence added because of an underrun.
+  CheckIsAudible(Span(aBuffer, samplesRead), mOutputChannels);
+
+  return SampleToFrame(samplesRead);
 }
 
 bool AudioSink::Ended() const {
   // Return true when error encountered so AudioStream can start draining.
-  return mProcessedQueue.IsFinished() || mErrored;
+  // Both atomic so we don't need locking
+  return mProcessedQueueFinished || mErrored;
 }
 
-void AudioSink::CheckIsAudible(const AudioData* aData) {
-  MOZ_ASSERT(aData);
-
-  mAudibilityMonitor.Process(aData);
+void AudioSink::CheckIsAudible(const Span<AudioDataValue>& aInterleaved,
+                               size_t aChannel) {
+  mAudibilityMonitor.ProcessInterleaved(aInterleaved, aChannel);
   bool isAudible = mAudibilityMonitor.RecentlyAudible();
 
   if (isAudible != mIsAudioDataAudible) {
     mIsAudioDataAudible = isAudible;
+    SINK_LOG("Notifying that audio is now %s",
+             mIsAudioDataAudible ? "audible" : "inaudible");
     mAudibleEvent.Notify(mIsAudioDataAudible);
   }
 }
 
-void AudioSink::OnAudioPopped(const RefPtr<AudioData>& aSample) {
+void AudioSink::OnAudioPopped() {
   SINK_LOG_V("AudioStream has used an audio packet.");
   NotifyAudioNeeded();
 }
@@ -352,15 +399,35 @@ void AudioSink::OnAudioPushed(const RefPtr<AudioData>& aSample) {
   NotifyAudioNeeded();
 }
 
+uint32_t AudioSink::AudioQueuedInRingBufferMS() const {
+  return static_cast<uint32_t>(
+      1000 * SampleToFrame(mProcessedSPSCQueue->AvailableRead()) / mOutputRate);
+}
+
+uint32_t AudioSink::SampleToFrame(uint32_t aSamples) const {
+  return aSamples / mOutputChannels;
+}
+
 void AudioSink::NotifyAudioNeeded() {
   MOZ_ASSERT(mOwnerThread->IsCurrentThreadIn(),
              "Not called from the owner's thread");
 
-  // Always ensure we have two processed frames pending to allow for processing
-  // latency.
   while (mAudioQueue.GetSize() &&
-         (mAudioQueue.IsFinished() || mProcessedQueueLength < LOW_AUDIO_USECS ||
-          mProcessedQueue.GetSize() < 2)) {
+         AudioQueuedInRingBufferMS() <
+             static_cast<uint32_t>(mProcessedQueueThresholdMS)) {
+    // Check if there's room in our ring buffer.
+    if (mAudioQueue.PeekFront()->Frames() >
+        SampleToFrame(mProcessedSPSCQueue->AvailableWrite())) {
+      SINK_LOG_V("Can't push %u frames. In ringbuffer %u / %u\n",
+                 mAudioQueue.PeekFront()->Frames(),
+                 SampleToFrame(mProcessedSPSCQueue->AvailableRead()),
+                 SampleToFrame(mProcessedSPSCQueue->Capacity()));
+      return;
+    }
+    SINK_LOG_V("Pushing %u frames. In ringbuffer %u / %u\n",
+               mAudioQueue.PeekFront()->Frames(),
+               SampleToFrame(mProcessedSPSCQueue->AvailableRead()),
+               SampleToFrame(mProcessedSPSCQueue->Capacity()));
     RefPtr<AudioData> data = mAudioQueue.PopFront();
 
     // Ignore the element with 0 frames and try next.
@@ -376,7 +443,7 @@ void AudioSink::NotifyAudioNeeded() {
                  mConverter ? mConverter->InputConfig().Rate() : 0,
                  data->mChannels, data->mRate);
 
-      DrainConverter();
+      DrainConverter(SampleToFrame(mProcessedSPSCQueue->AvailableWrite()));
 
       // mFramesParsed indicates the current playtime in frames at the current
       // input sampling rate. Recalculate it per the new sampling rate.
@@ -432,10 +499,15 @@ void AudioSink::NotifyAudioNeeded() {
       // The next audio packet begins some time after the end of the last packet
       // we pushed to the audio hardware. We must push silence into the audio
       // hardware so that the next audio packet begins playback at the correct
-      // time.
-      missingFrames = std::min<int64_t>(INT32_MAX, missingFrames.value());
+      // time. But don't push more than the ring buffer can receive.
+      missingFrames = std::min<int64_t>(
+          std::min<int64_t>(INT32_MAX, missingFrames.value()),
+          SampleToFrame(mProcessedSPSCQueue->AvailableWrite()));
       mFramesParsed += missingFrames.value();
 
+      SINK_LOG("Gap in the audio input, push %" PRId64 " frames of silence",
+               missingFrames.value());
+
       RefPtr<AudioData> silenceData;
       AlignedAudioBuffer silenceBuffer(missingFrames.value() * data->mChannels);
       if (!silenceBuffer) {
@@ -451,6 +523,7 @@ void AudioSink::NotifyAudioNeeded() {
       } else {
         silenceData = CreateAudioFromBuffer(std::move(silenceBuffer), data);
       }
+      TRACE("Pushing silence");
       PushProcessedAudio(silenceData);
     }
 
@@ -468,10 +541,10 @@ void AudioSink::NotifyAudioNeeded() {
     }
   }
 
-  if (mAudioQueue.IsFinished()) {
+  if (mAudioQueue.IsFinished() && mAudioQueue.GetSize() == 0) {
     // We have reached the end of the data, drain the resampler.
-    DrainConverter();
-    mProcessedQueue.Finish();
+    DrainConverter(SampleToFrame(mProcessedSPSCQueue->AvailableWrite()));
+    mProcessedQueueFinished = true;
   }
 }
 
@@ -479,14 +552,22 @@ uint32_t AudioSink::PushProcessedAudio(AudioData* aData) {
   if (!aData || !aData->Frames()) {
     return 0;
   }
-  mProcessedQueue.Push(aData);
-  mProcessedQueueLength += FramesToUsecs(aData->Frames(), mOutputRate).value();
+  int framesToEnqueue = static_cast<int>(aData->Frames() * aData->mChannels);
+  TRACE_COMMENT("AudioSink::PushProcessedAudio", "%u frames (%u/%u)",
+                framesToEnqueue,
+                SampleToFrame(mProcessedSPSCQueue->AvailableWrite()),
+                SampleToFrame(mProcessedSPSCQueue->Capacity()));
+  DebugOnly<int> rv =
+      mProcessedSPSCQueue->Enqueue(aData->Data().Elements(), framesToEnqueue);
+  NS_WARNING_ASSERTION(
+      rv == static_cast<int>(aData->Frames() * aData->mChannels),
+      "AudioSink ring buffer over-run, can't push new data");
   return aData->Frames();
 }
 
 already_AddRefed<AudioData> AudioSink::CreateAudioFromBuffer(
     AlignedAudioBuffer&& aBuffer, AudioData* aReference) {
-  uint32_t frames = aBuffer.Length() / mOutputChannels;
+  uint32_t frames = SampleToFrame(aBuffer.Length());
   if (!frames) {
     return nullptr;
   }
@@ -518,7 +599,7 @@ uint32_t AudioSink::DrainConverter(uint32_t aMaxFrames) {
   AlignedAudioBuffer convertedData =
       mConverter->Process(AudioSampleBuffer(AlignedAudioBuffer())).Forget();
 
-  uint32_t frames = convertedData.Length() / mOutputChannels;
+  uint32_t frames = SampleToFrame(convertedData.Length());
   if (!convertedData.SetLength(std::min(frames, aMaxFrames) *
                                mOutputChannels)) {
     // This can never happen as we were reducing the length of convertData.
@@ -528,11 +609,7 @@ uint32_t AudioSink::DrainConverter(uint32_t aMaxFrames) {
 
   RefPtr<AudioData> data =
       CreateAudioFromBuffer(std::move(convertedData), lastPacket);
-  if (!data) {
-    return 0;
-  }
-  mProcessedQueue.Push(data);
-  return data->Frames();
+  return PushProcessedAudio(data);
 }
 
 void AudioSink::GetDebugInfo(dom::MediaSinkDebugInfo& aInfo) {
