@@ -11,6 +11,7 @@
 #include "EMEDecoderModule.h"
 #include "GLImages.h"
 #include "JavaCallbacksSupport.h"
+#include "MediaCodec.h"
 #include "MediaData.h"
 #include "MediaInfo.h"
 #include "SimpleMap.h"
@@ -60,7 +61,7 @@ class RenderOrReleaseOutput {
   java::Sample::GlobalRef mSample;
 };
 
-class RemoteVideoDecoder : public RemoteDataDecoder {
+class RemoteVideoDecoder final : public RemoteDataDecoder {
  public:
   // Render the output to the surface when the frame is sent
   // to compositor, or release it if not presented.
@@ -131,8 +132,9 @@ class RemoteVideoDecoder : public RemoteDataDecoder {
 
   RefPtr<InitPromise> Init() override {
     mThread = GetCurrentSerialEventTarget();
-    java::sdk::BufferInfo::LocalRef bufferInfo;
-    if (NS_FAILED(java::sdk::BufferInfo::New(&bufferInfo)) || !bufferInfo) {
+    java::sdk::MediaCodec::BufferInfo::LocalRef bufferInfo;
+    if (NS_FAILED(java::sdk::MediaCodec::BufferInfo::New(&bufferInfo)) ||
+        !bufferInfo) {
       return InitPromise::CreateAndReject(NS_ERROR_OUT_OF_MEMORY, __func__);
     }
     mInputBufferInfo = bufferInfo;
@@ -183,6 +185,11 @@ class RemoteVideoDecoder : public RemoteDataDecoder {
       MediaRawData* aSample) override {
     AssertOnThread();
 
+    if (NeedsNewDecoder()) {
+      return DecodePromise::CreateAndReject(NS_ERROR_DOM_MEDIA_NEED_NEW_DECODER,
+                                            __func__);
+    }
+
     const VideoInfo* config =
         aSample->mTrackInfo ? aSample->mTrackInfo->GetAsVideoInfo() : &mConfig;
     MOZ_ASSERT(config);
@@ -265,7 +272,18 @@ class RemoteVideoDecoder : public RemoteDataDecoder {
     UniquePtr<layers::SurfaceTextureImage::SetCurrentCallback> releaseSample(
         new CompositeListener(mJavaDecoder, aSample));
 
-    java::sdk::BufferInfo::LocalRef info = aSample->Info();
+    // If our output surface has been released (due to the GPU process crashing)
+    // then request a new decoder, which will in turn allocate a new
+    // Surface. This is usually be handled by the Error() callback, but on some
+    // devices (or at least on the emulator) the java decoder does not raise an
+    // error when the Surface is released. So we raise this error here as well.
+    if (NeedsNewDecoder()) {
+      Error(MediaResult(NS_ERROR_DOM_MEDIA_NEED_NEW_DECODER,
+                        RESULT_DETAIL("VideoCallBack::HandleOutput")));
+      return;
+    }
+
+    java::sdk::MediaCodec::BufferInfo::LocalRef info = aSample->Info();
     MOZ_ASSERT(info);
 
     int32_t flags;
@@ -316,6 +334,10 @@ class RemoteVideoDecoder : public RemoteDataDecoder {
     }
   }
 
+  bool NeedsNewDecoder() const override {
+    return !mSurface || mSurface->IsReleased();
+  }
+
   const VideoInfo mConfig;
   java::GeckoSurface::GlobalRef mSurface;
   AndroidSurfaceTextureHandle mSurfaceHandle;
@@ -331,7 +353,7 @@ class RemoteVideoDecoder : public RemoteDataDecoder {
   Maybe<TimeUnit> mLatestOutputTime;
 };
 
-class RemoteAudioDecoder : public RemoteDataDecoder {
+class RemoteAudioDecoder final : public RemoteDataDecoder {
  public:
   RemoteAudioDecoder(const AudioInfo& aConfig,
                      java::sdk::MediaFormat::Param aFormat,
@@ -343,18 +365,24 @@ class RemoteAudioDecoder : public RemoteDataDecoder {
     bool formatHasCSD = false;
     NS_ENSURE_SUCCESS_VOID(aFormat->ContainsKey(u"csd-0"_ns, &formatHasCSD));
 
-    if (!formatHasCSD && aConfig.mCodecSpecificConfig->Length() >= 2) {
+    // It would be nice to instead use more specific information here, but
+    // we force a byte buffer for now since this handles arbitrary codecs.
+    // TODO(bug 1768564): implement further type checking for codec data.
+    RefPtr<MediaByteBuffer> audioCodecSpecificBinaryBlob =
+        ForceGetAudioCodecSpecificBlob(aConfig.mCodecSpecificConfig);
+    if (!formatHasCSD && audioCodecSpecificBinaryBlob->Length() >= 2) {
       jni::ByteBuffer::LocalRef buffer(env);
-      buffer = jni::ByteBuffer::New(aConfig.mCodecSpecificConfig->Elements(),
-                                    aConfig.mCodecSpecificConfig->Length());
+      buffer = jni::ByteBuffer::New(audioCodecSpecificBinaryBlob->Elements(),
+                                    audioCodecSpecificBinaryBlob->Length());
       NS_ENSURE_SUCCESS_VOID(aFormat->SetByteBuffer(u"csd-0"_ns, buffer));
     }
   }
 
   RefPtr<InitPromise> Init() override {
     mThread = GetCurrentSerialEventTarget();
-    java::sdk::BufferInfo::LocalRef bufferInfo;
-    if (NS_FAILED(java::sdk::BufferInfo::New(&bufferInfo)) || !bufferInfo) {
+    java::sdk::MediaCodec::BufferInfo::LocalRef bufferInfo;
+    if (NS_FAILED(java::sdk::MediaCodec::BufferInfo::New(&bufferInfo)) ||
+        !bufferInfo) {
       return InitPromise::CreateAndReject(NS_ERROR_OUT_OF_MEMORY, __func__);
     }
     mInputBufferInfo = bufferInfo;
@@ -482,7 +510,7 @@ class RemoteAudioDecoder : public RemoteDataDecoder {
 
     RenderOrReleaseOutput autoRelease(mJavaDecoder, aSample);
 
-    java::sdk::BufferInfo::LocalRef info = aSample->Info();
+    java::sdk::MediaCodec::BufferInfo::LocalRef info = aSample->Info();
     MOZ_ASSERT(info);
 
     int32_t flags = 0;
@@ -661,17 +689,24 @@ RefPtr<ShutdownPromise> RemoteDataDecoder::Shutdown() {
   return ShutdownPromise::CreateAndResolve(true, __func__);
 }
 
-static java::sdk::CryptoInfo::LocalRef GetCryptoInfoFromSample(
-    const MediaRawData* aSample) {
+using CryptoInfoResult =
+    Result<java::sdk::MediaCodec::CryptoInfo::LocalRef, nsresult>;
+
+static CryptoInfoResult GetCryptoInfoFromSample(const MediaRawData* aSample) {
   auto& cryptoObj = aSample->mCrypto;
+  java::sdk::MediaCodec::CryptoInfo::LocalRef cryptoInfo;
 
   if (!cryptoObj.IsEncrypted()) {
-    return nullptr;
+    return CryptoInfoResult(cryptoInfo);
   }
 
-  java::sdk::CryptoInfo::LocalRef cryptoInfo;
-  nsresult rv = java::sdk::CryptoInfo::New(&cryptoInfo);
-  NS_ENSURE_SUCCESS(rv, nullptr);
+  static bool supportsCBCS = java::CodecProxy::SupportsCBCS();
+  if (cryptoObj.mCryptoScheme == CryptoScheme::Cbcs && !supportsCBCS) {
+    return CryptoInfoResult(NS_ERROR_DOM_MEDIA_NOT_SUPPORTED_ERR);
+  }
+
+  nsresult rv = java::sdk::MediaCodec::CryptoInfo::New(&cryptoInfo);
+  NS_ENSURE_SUCCESS(rv, CryptoInfoResult(rv));
 
   uint32_t numSubSamples = std::min<uint32_t>(
       cryptoObj.mPlainSizes.Length(), cryptoObj.mEncryptedSizes.Length());
@@ -689,7 +724,7 @@ static java::sdk::CryptoInfo::LocalRef GetCryptoInfoFromSample(
   uint32_t codecSpecificDataSize = aSample->Size() - totalSubSamplesSize;
   // Size of codec specific data("CSD") for Android java::sdk::MediaCodec usage
   // should be included in the 1st plain size if it exists.
-  if (!plainSizes.IsEmpty()) {
+  if (codecSpecificDataSize > 0 && !plainSizes.IsEmpty()) {
     // This shouldn't overflow as the the plain size should be UINT16_MAX at
     // most, and the CSD should never be that large. Checked int acts like a
     // diagnostic assert here to help catch if we ever have insane inputs.
@@ -699,9 +734,26 @@ static java::sdk::CryptoInfo::LocalRef GetCryptoInfoFromSample(
   }
 
   static const int kExpectedIVLength = 16;
-  auto tempIV(cryptoObj.mIV);
+  nsTArray<uint8_t> tempIV(kExpectedIVLength);
+  jint mode;
+  switch (cryptoObj.mCryptoScheme) {
+    case CryptoScheme::None:
+      mode = java::sdk::MediaCodec::CRYPTO_MODE_UNENCRYPTED;
+      MOZ_ASSERT(cryptoObj.mIV.Length() <= kExpectedIVLength);
+      tempIV.AppendElements(cryptoObj.mIV);
+      break;
+    case CryptoScheme::Cenc:
+      mode = java::sdk::MediaCodec::CRYPTO_MODE_AES_CTR;
+      MOZ_ASSERT(cryptoObj.mIV.Length() <= kExpectedIVLength);
+      tempIV.AppendElements(cryptoObj.mIV);
+      break;
+    case CryptoScheme::Cbcs:
+      mode = java::sdk::MediaCodec::CRYPTO_MODE_AES_CBC;
+      MOZ_ASSERT(cryptoObj.mConstantIV.Length() <= kExpectedIVLength);
+      tempIV.AppendElements(cryptoObj.mConstantIV);
+      break;
+  }
   auto tempIVLength = tempIV.Length();
-  MOZ_ASSERT(tempIVLength <= kExpectedIVLength);
   for (size_t i = tempIVLength; i < kExpectedIVLength; i++) {
     // Padding with 0
     tempIV.AppendElement(0);
@@ -710,10 +762,13 @@ static java::sdk::CryptoInfo::LocalRef GetCryptoInfoFromSample(
   cryptoInfo->Set(numSubSamples, mozilla::jni::IntArray::From(plainSizes),
                   mozilla::jni::IntArray::From(cryptoObj.mEncryptedSizes),
                   mozilla::jni::ByteArray::From(cryptoObj.mKeyId),
-                  mozilla::jni::ByteArray::From(tempIV),
-                  java::sdk::MediaCodec::CRYPTO_MODE_AES_CTR);
+                  mozilla::jni::ByteArray::From(tempIV), mode);
+  if (mode == java::sdk::MediaCodec::CRYPTO_MODE_AES_CBC) {
+    java::CodecProxy::SetCryptoPatternIfNeeded(
+        cryptoInfo, cryptoObj.mCryptByteBlock, cryptoObj.mSkipByteBlock);
+  }
 
-  return cryptoInfo;
+  return CryptoInfoResult(cryptoInfo);
 }
 
 RefPtr<MediaDataDecoder::DecodePromise> RemoteDataDecoder::Decode(
@@ -726,8 +781,13 @@ RefPtr<MediaDataDecoder::DecodePromise> RemoteDataDecoder::Decode(
 
   SetState(State::DRAINABLE);
   mInputBufferInfo->Set(0, aSample->Size(), aSample->mTime.ToMicroseconds(), 0);
-  int64_t session = mJavaDecoder->Input(bytes, mInputBufferInfo,
-                                        GetCryptoInfoFromSample(aSample));
+  CryptoInfoResult crypto = GetCryptoInfoFromSample(aSample);
+  if (crypto.isErr()) {
+    return DecodePromise::CreateAndReject(
+        MediaResult(crypto.unwrapErr(), __func__), __func__);
+  }
+  int64_t session =
+      mJavaDecoder->Input(bytes, mInputBufferInfo, crypto.unwrap());
   if (session == java::CodecProxy::INVALID_SESSION) {
     return DecodePromise::CreateAndReject(
         MediaResult(NS_ERROR_OUT_OF_MEMORY, __func__), __func__);
@@ -832,8 +892,17 @@ void RemoteDataDecoder::Error(const MediaResult& aError) {
   if (GetState() == State::SHUTDOWN) {
     return;
   }
-  mDecodePromise.RejectIfExists(aError, __func__);
-  mDrainPromise.RejectIfExists(aError, __func__);
+
+  // If we know we need a new decoder (eg because RemoteVideoDecoder's mSurface
+  // has been released due to a GPU process crash) then override the error to
+  // request a new decoder.
+  const MediaResult& error =
+      NeedsNewDecoder()
+          ? MediaResult(NS_ERROR_DOM_MEDIA_NEED_NEW_DECODER, __func__)
+          : aError;
+
+  mDecodePromise.RejectIfExists(error, __func__);
+  mDrainPromise.RejectIfExists(error, __func__);
 }
 
 }  // namespace mozilla
