@@ -21,6 +21,8 @@ class MediaDataDecoderProxy
     : public MediaDataDecoder,
       public DecoderDoctorLifeLogger<MediaDataDecoderProxy> {
  public:
+  NS_INLINE_DECL_THREADSAFE_REFCOUNTING(MediaDataDecoderProxy, final);
+
   explicit MediaDataDecoderProxy(
       already_AddRefed<MediaDataDecoder> aProxyDecoder,
       already_AddRefed<nsISerialEventTarget> aProxyThread = nullptr)
@@ -42,6 +44,9 @@ class MediaDataDecoderProxy
   bool SupportDecoderRecycling() const override;
   ConversionRequired NeedsConversion() const override;
 
+ protected:
+  ~MediaDataDecoderProxy() = default;
+
  private:
   // Set on construction and clear on the proxy thread if set.
   RefPtr<MediaDataDecoder> mProxyDecoder;
